"""
Module handling the injection of predictions
back into root ntuples.
"""
import collections
import os
from pickle import load
from array import array
import tensorflow as tf
import numpy as np
import uproot
import ROOT
import HelperModules.HelperFunctions as hf
from HelperModules.Settings import Settings
from HelperModules.MessageHandler import ErrorMessage


def load_data(f, t, inputs):
    Events = uproot.open(f + ":" + t)
    try:
        e = Events.arrays(inputs, library="pd")
    except uproot.exceptions.KeyInFileError as e:
        ErrorMessage("The variable '%s' is missing in tree '%s' in file '%s'. Consider using the 'IgnoreTreenames' option." % (e.args[0], t, f))
    return e

def do_scaling(df_X, scaler, nFolds):
    df_X_Folds = [df_X[df_X["eventNumber"] % nFolds == i] for i in range(nFolds)]
    # Let's reorder the event numbers and concatenate them again. We need this for the keys of our prediction dictionary
    df_X_Folds = [X.drop(["eventNumber"], axis=1) for X in df_X_Folds]
    # We need to make sure we actually have values to pass to the scaler. Otherwise this will crash
    if scaler is None:
        return [X.values if len(X.values)!=0 else [] for X in df_X_Folds]
    return [scaler.transform(X.values) if len(X.values)!=0 else [] for X in df_X_Folds]

def predict_folds(models, x_in, e_numbers, nFolds, is_bdt=False, is_binary=False):
    if is_bdt:
        if is_binary:
            P = [models[i].predict_proba(x_in[i])[:, 1] for i in range(nFolds) if len(x_in[i])!=0]  # scikit-learn provides both [1-P,P]. We only want P.
        else:
            P = [models[i].predict_proba(x_in[i]) for i in range(nFolds) if len(x_in[i])!=0]
    else:
        P = [models[i].predict(x_in[i]) for i in range(nFolds) if len(x_in[i])!=0]
    e_array = [[] for i in range(nFolds)]
    e_array_ind = [[] for i in range(nFolds)]
    for ind, e_number in enumerate(e_numbers):
        e_array[e_number % nFolds].append(e_number)
        e_array_ind[e_number % nFolds].append(ind)
    e_array_trunc = np.concatenate(e_array).astype(int)
    e_array_ind_trunc = np.concatenate(e_array_ind).astype(int)
    P_trunc = np.concatenate(P).astype(float)
    P_array = [None] * len(e_array_trunc)
    for ind, val in zip(e_array_ind_trunc, P_trunc):
        P_array[ind] = val
    return P_array

def do_PreCalculation(fName, treename, In_dict):
    """
    Function performing the pre-calculation, i.e. calculating
    the predictions given by the model
    keyword arguments:
    fName --- file name (string)
    treename --- treename (string)
    """
    Predictions, column_names = [], []
    for ParameterDict in In_dict.values():
        X_raw_df = load_data(f=fName, t=treename, inputs=ParameterDict["ModelInputs"])
        X_Scaled = do_scaling(X_raw_df, ParameterDict["Scaler"], ParameterDict["ModelFolds"])
        Predictions.append(predict_folds(models=ParameterDict["ModelObjects"],
                                         x_in=X_Scaled,
                                         e_numbers=X_raw_df["eventNumber"],
                                         nFolds=ParameterDict["ModelFolds"],
                                         is_bdt=ParameterDict["Model"].isClassificationBDT(),
                                         is_binary=(len(ParameterDict["OutputLabels"])==1)))
        column_names.append([ParameterDict["Model"].get("Name") + "_" + cl for cl in ParameterDict["OutputLabels"]])
    c_dicts = [dict(zip(np.concatenate(column_names), p)) for p in np.concatenate(Predictions, 1)]
    del column_names, Predictions
    return c_dicts


class Injector(object):
    """
    Class handling the injection of ntuples into root files.
    Takes args from argparse as parameter.
    """

    def __init__(self, args):
        self.m_args = args
        self.m_CfgSettings = [Settings(ConfigPath, option="all") for ConfigPath in args.configfile]

    def check_modelnames(self):
        ModelNames = [Model.get("Name") for Model in [s.get_Model() for s in self.m_CfgSettings]]
        if len(ModelNames) != len(set(ModelNames)):
            Duplicates = [item for item, count in collections.Counter(ModelNames).items() if count > 1]
            ErrorMessage("You are using models that have identical `Model Names`. The following names occur multiple times %s." % Duplicates)
        return

    def get_injection_dict(self):
        ModelNames = [Model.get("Name") for Model in [s.get_Model() for s in self.m_CfgSettings]]
        InjectionDict = {ModelName: {} for ModelName in ModelNames}
        for S, ModelPath in zip(self.m_CfgSettings, self.m_args.model_path):
            Model = S.get_Model()
            nFolds = S.get_General().get("Folds")
            ModelInputs = S.get_VarNames().copy()
            ModelInputs += ["eventNumber"]
            if nFolds != 1:
                Folds = ["_Fold" + str(i) for i in range(nFolds)]
            else:
                Folds = [""]
            # Getting the paths of the models and the objects themselves
            if Model.isClassificationBDT():
                InjectionDict[Model.get("Name")]["ModelPaths"] = [hf.ensure_trailing_slash(ModelPath) + Model.get("Name") + "_" + Model.get("Type") + Fold + ".pkl" for Fold in Folds]
                ModelObjects = []
                for path in InjectionDict[Model.get("Name")]["ModelPaths"]:
                    with open(path, 'rb') as f:
                        ModelObjects.append(load(f))
                InjectionDict[Model.get("Name")]["ModelObjects"] = ModelObjects
            elif Model.isClassificationDNN() or Model.isReconstruction() or Model.isRegression():
                InjectionDict[Model.get("Name")]["ModelPaths"] = [hf.ensure_trailing_slash(ModelPath) + Model.get("Name") + "_" + Model.get("Type") + Fold + ".h5" for Fold in Folds]
                ModelObjects = [tf.keras.models.load_model(path) for path in InjectionDict[Model.get("Name")]["ModelPaths"]]
                InjectionDict[Model.get("Name")]["ModelObjects"] = ModelObjects
            # Getting additional parameters
            InjectionDict[Model.get("Name")]["Scaler"] = load(open(hf.ensure_trailing_slash(ModelPath) + "scaler.pkl", 'rb'))
            InjectionDict[Model.get("Name")]["OutputLabels"] = [classlabel.replace(" ", "") for classlabel in Model.get("OutputLabels")]
            InjectionDict[Model.get("Name")]["ModelInputs"] = ModelInputs
            InjectionDict[Model.get("Name")]["ModelFolds"] = nFolds
            InjectionDict[Model.get("Name")]["Model"] = Model
            InjectionDict[Model.get("Name")]["Type"] = Model.get("Type")
        return InjectionDict

    def process_tree(self, in_File, out_File, tName, fName, In_dict):
        Tree = in_File.Get(tName)
        if tName not in self.m_args.tree_wildcard and self.m_args.tree_wildcard != "":
            Tree.Delete()
            return
        # black listing a few trees which are simply copied
        if any([tName in c.get_General().get("IgnoreTreenames") for c in self.m_CfgSettings]):
            NewTree = Tree.CloneTree(-1, "fast")
            out_File.Write()
            Tree.Delete()
            return
        NewTree = Tree.CloneTree(0)
        NewBranches = {}
        for ModelName, ParameterDict in In_dict.items():
            for Label in ParameterDict["OutputLabels"]:
                NewBranches[ModelName + "_" + Label] = array('f', [0])
        for BranchName, NewBranch in NewBranches.items():
            NewTree.Branch(BranchName, NewBranch, BranchName + "/F")
        if Tree.GetEntries() == 0:
            out_File.Write()
            Tree.Delete()
            return
        event_dicts = do_PreCalculation(fName=fName, treename=tName, In_dict=In_dict)
        for entry_number, event_dict in zip(range(0, Tree.GetEntries()), event_dicts):
            Tree.GetEntry(entry_number)
            for name in NewBranches:
                NewBranches[name][0] = event_dict[name]
            NewTree.Fill()
        event_dict.clear()
        NewBranches.clear()
        Tree.Delete()
        out_File.Write()

    def do_Injection(self, fName):
        """
        Function to perform the injection of ntuples into the actual root file.

        Keyordword arguments:
        fName --- file name (string)
        """
        # Create a new root file
        File = ROOT.TFile(fName)
        NewFile = ROOT.TFile(fName.replace(os.path.abspath(self.m_args.input_path), os.path.abspath(self.m_args.output_path)), "recreate")
        # Loop over trees in root file
        keys = File.GetListOfKeys()
        Treenames = []
        for key in keys:
            if key.GetClassName() != "TTree":  # ignore TObjects that are not TTrees
                continue
            if key.ReadObj().GetName() == "":  # ignore TTrees that do not have a name
                continue
            Treenames.append(key.ReadObj().GetName())
        for tName in Treenames:
            self.process_tree(in_File=File,
                              out_File=NewFile,
                              tName=tName,
                              fName=fName,
                              In_dict=self.get_injection_dict())
        del Treenames
        NewFile.Close()
        NewFile.Delete()
        File.Close()
        File.Delete()
